
/**************************************************************************
**
** Copyright (C) EMC Corporation 2001, All rights reserved.
**
** FILE NAME: ClientFileSysManager.cpp
**
** DESCRIPTION: This class mainly manage all file system information used
**              by one client.
**              
**
** AUTHOR: Yubing Wang       Mar-02-2001
**
** REVISION HISTORY:
** 
**  Who         Date       Description of Change
** ---------------------------------------------------------------------
**
** (This is only for the major code change that does not include bug fixes.
** Comments for bug fixes should be placed directly where the bugs are.)
**
****************************************************************************
**
**
** FUNCTION LIST:
**     CClientFileSysManager::CClientFileSysManager()
**     CClientFileSysManager::~CClientFileSysManager()
**
**
****************************************************************************/


#include "Common.h"
#include "ClientFileSysManager.h"
#include "ClientLoadMgrThread.h"

/***************************************************************************/
/*           the constant for read and write size distribution             */

int CClientFileSysManager::m_BlockSize  = 8192;

SIOSizeDist CClientFileSysManager::m_ReadDist =
{
    {   0,  0, 0 },     /*   0%    1 -   7 KB	*/
    {  85,  1, 0 },     /*  85%    8       KB	*/
    {  93,  2, 1 },     /*   8%   17 -  23 KB	*/
    {  97,  4, 1 },     /*   4%   33 -  39 KB	*/
    {  99,  8, 1 },     /*   2%   65 -  71 KB	*/
    { 100, 16, 1 },     /*   1%  129 - 135 KB	*/
};

SIOSizeDist CClientFileSysManager::m_WriteDist =
{
    {  49,  0, 1 },     /*  49%    1 -   7 KB	*/
    {  85,  1, 1 },     /*  36%    9 -  15 KB	*/
    {  93,  2, 1 },     /*   8%   17 -  23 KB	*/
    {  97,  4, 1 },     /*   4%   33 -  39 KB	*/
    {  99,  8, 1 },     /*   2%   65 -  71 KB	*/
    { 100, 16, 1 },     /*   1%  129 - 135 KB	*/
};

int CClientFileSysManager::m_MaxBufNum = 17;

double CClientFileSysManager::m_AvgOpsPerReadReq = 1.64;

double CClientFileSysManager::m_AvgOpsPerWriteReq = 2.0;

/***************************************************************************/

CClientFileSysManager::CClientFileSysManager(CClientLoadMgrThread *pLoadMgrThread)
{
    m_num = 0;
    m_LoadMgrThread = pLoadMgrThread;
}

CClientFileSysManager::~CClientFileSysManager()
{
    DeleteAllFileSysBase();
    DeleteAllFileSysInfo();
}

void CClientFileSysManager::AddFileSysBase(CClientFileSysBase *fsbase, int prob)
{
    _ASSERT(m_num < FILESYS_MAXNUM);
    m_fsbase[m_num] = fsbase;
    if (m_num == 0)
        m_range[m_num] = prob;
    else
        m_range[m_num] = m_range[m_num-1] + prob;
    m_fsinfo[m_num] = NULL;
    m_num ++;
}

int CClientFileSysManager::GetFileSysNum()
{
    return m_num;
}

CClientFileSysBase *CClientFileSysManager::GetFileSysBase(int fs)
{
    return m_fsbase[fs];
}

void CClientFileSysManager::DeleteAllFileSysBase()
{
    for (int i=0; i<m_num; i++)
    {
        if (m_fsbase[i] != NULL)
            delete m_fsbase[i];
        m_fsbase[i] = NULL;

        if (m_fsinfo[i] != NULL)
            delete m_fsinfo[i];
        m_fsinfo[i] = NULL;
    }
    m_num = 0;
}

void CClientFileSysManager::SetFileSysInfo(int fs, CClientFileSysInfo *fsinfo)
{
    _ASSERT(fs < m_num);
    if (m_fsinfo[fs] != NULL)
        delete m_fsinfo[fs];
    m_fsinfo[fs] = fsinfo;
}

CClientFileSysInfo *CClientFileSysManager::GetFileSysInfo(int fs)
{
    return m_fsinfo[fs];
}

void CClientFileSysManager::DeleteAllFileSysInfo()
{
    for (int i=0; i<m_num; i++)
    {
        if (m_fsinfo[i] != NULL)
            delete m_fsinfo[i];
        m_fsinfo[i] = NULL;
    }
}

int CClientFileSysManager::GetBlockSize()
{
    return m_BlockSize;
}

double CClientFileSysManager::GetAvgOpsPerReadReq()
{
    return m_AvgOpsPerReadReq;
}

double CClientFileSysManager::GetAvgOpsPerWriteReq()
{
    return m_AvgOpsPerWriteReq;
}

/* op=1 for read, op=2 for write */
int CClientFileSysManager::GetIOOpSize(int op)
{
    SIOSizeDist *dist = (op==1) ? &m_ReadDist : &m_WriteDist;

    /* determine number of full buffers and their total size */
    int ratio = GetRandom() % 100;
    int i = 0;
    while ((*dist)[i].pcnt <= ratio)
        i ++;
    int size = (*dist)[i].bufs * m_BlockSize;
    if ((*dist)[i].frags)
    {
        /* determine size of fragment: 1KB ~ (Kb_per_block - 1) KB */
        ratio = GetRandom() % 100;
        int Kb_per_block = m_BlockSize / 1024;
        if (Kb_per_block > 1)
            ratio = ratio % (Kb_per_block-1);
        else
            ratio = 0;
        size += (ratio + 1) * 1024;
    }

    return size;
}

int CClientFileSysManager::SelectFileSys()
{
    int value = GetRandom() % m_range[m_num-1];
    int i;
    for (i=0; i<m_num; i++)
    {
        if (value < m_range[i])
            break;
    }
    return i;
}

BOOL CClientFileSysManager::GetIOFileInfo(int req_size, char *name, int &fs, int &handle, int &size)
{
    fs = SelectFileSys();
    return m_fsinfo[fs]->GetIOFileInfo(req_size, name, handle, size);
}

void CClientFileSysManager::SetIOFileInfo(int fs, int handle, int size)
{
    m_fsinfo[fs]->SetIOFileInfo(handle, size);
}

/* option: 0(access) 1(remove) 2(create) */
BOOL CClientFileSysManager::GetNonIOFileInfo(int option, char *name, int &fs, int &handle, int &mode, int ref_fs)
{
    fs = (ref_fs != -1) ? ref_fs : SelectFileSys();
    return m_fsinfo[fs]->GetNonIOFileInfo(option, name, handle, mode);
}

/* option: 0(access) 1(remove) 2(create) */
void CClientFileSysManager::SetNonIOFileInfo(int option, int fs, int handle, int mode)
{
    m_fsinfo[fs]->SetNonIOFileInfo(option, handle, mode);
}

/* option: 0(access) 1(remove) 2(create) */
BOOL CClientFileSysManager::GetDirInfo(int option, char *name, int &fs, int &handle)
{
    fs = SelectFileSys();
    return m_fsinfo[fs]->GetDirInfo(option, name, handle);
}

/* option: 0(access) 1(remove) 2(create) */
void CClientFileSysManager::SetDirInfo(int option, int fs, int handle)
{
    m_fsinfo[fs]->SetDirInfo(option, handle);
}

/* option: 0(access) 1(remove) 2(create) */
BOOL CClientFileSysManager::GetSymlinkInfo(int option, char *name, int &fs, int &handle)
{
    fs = SelectFileSys();
    return m_fsinfo[fs]->GetSymlinkInfo(option, name, handle);
}

/* option: 0(access) 1(remove) 2(create) */
void CClientFileSysManager::SetSymlinkInfo(int option, int fs, int handle)
{
    m_fsinfo[fs]->SetSymlinkInfo(option, handle);
}
