
/**************************************************************************
**
** Copyright (C) EMC Corporation 2001, All rights reserved.
**
** FILE NAME: ClientFileSysRefThread.cpp
**
** DESCRIPTION: This class mainly manage the thread to refresh file system.
**              
**
** AUTHOR: Yubing Wang       Mar-02-2001
**
** REVISION HISTORY:
** 
**  Who         Date       Description of Change
** ---------------------------------------------------------------------
**
** (This is only for the major code change that does not include bug fixes.
** Comments for bug fixes should be placed directly where the bugs are.)
**
****************************************************************************
**
**
** FUNCTION LIST:
**     CClientFileSysRefThread::CClientFileSysRefThread()
**     CClientFileSysRefThread::~CClientFileSysRefThread()
**     void CClientFileSysRefThread::run()
**
**
****************************************************************************/


#include <stdio.h>
#include <string.h>
#ifndef WIN32
#include <ctype.h>
#include <dirent.h>
#include <unistd.h>
#include <errno.h>
#endif
#include "Common.h"
#include "ClientFileSysRefThread.h"
#include "ClientFileSysRefPool.h"
#include "ClientFileSysBase.h"
#include "SafeFileOperation.h"


CClientFileSysRefThread::CClientFileSysRefThread()
{
    m_pool = NULL;
    m_fsbase = NULL;
#ifndef WIN32
    pthread_mutex_init(&m_lock, NULL);
    pthread_cond_init(&m_event, NULL);
    m_ready = FALSE;
#else
    m_readySem = CreateSemaphore(NULL, 0, 1, NULL);
#endif

    m_foundIO = NULL;
    m_foundFile = NULL;
    m_foundDir = NULL;
    m_foundSymlink = NULL;
}

CClientFileSysRefThread::CClientFileSysRefThread(int id, CClientFileSysRefPool *pool, CClientFileSysBase *fsbase)
{
    m_id = id;
    m_pool = pool;
    m_fsbase = fsbase;
#ifndef WIN32
    pthread_mutex_init(&m_lock, NULL);
    pthread_cond_init(&m_event, NULL);
    m_ready = FALSE;
#else
    m_readySem = CreateSemaphore(NULL, 0, 1, NULL);
#endif

    m_foundIO = NULL;
    m_foundFile = NULL;
    m_foundDir = NULL;
    m_foundSymlink = NULL;
}

CClientFileSysRefThread::~CClientFileSysRefThread()
{
#ifdef WIN32
    if (m_readySem != NULL)
        CloseHandle(m_readySem);
#endif

    if (m_foundIO != NULL)
        delete m_foundIO;
    if (m_foundFile != NULL)
        delete m_foundFile;
    if (m_foundDir != NULL)
        delete m_foundDir;
    if (m_foundSymlink != NULL)
        delete m_foundSymlink;
}

void CClientFileSysRefThread::run()
{
    while ( 1 )
    {
        /* wait for ready signal */
#ifndef WIN32
        LockMutex(m_lock);
        while (! m_ready)
            pthread_cond_wait(&m_event, &m_lock);
        m_ready = FALSE;
        UnlockMutex(m_lock);
#else
        WaitForSingleObject(m_readySem, INFINITE);
#endif

        /* check the quit signal */
        if (m_subdir < 0)
            break;

        /* do actual work */
        RefSubdir();

        /* resume to idle state */
        SetIdle();
    }
}

void CClientFileSysRefThread::SetFSBaseData(int *IONum, int **IOSize, int *FileNum, ushort **FileMode, int *DirNum, uchar **DirMode, int *SymlinkNum, uchar **SymlinkMode)
{
    m_IONum = IONum;
    m_IOSize = IOSize;
    m_FileNum = FileNum;
    m_FileMode = FileMode;
    m_DirNum = DirNum;
    m_DirMode = DirMode;
    m_SymlinkNum = SymlinkNum;
    m_SymlinkMode = SymlinkMode;
}

void CClientFileSysRefThread::SetFoundNum(int foundIO_Num, int foundFile_Num, int foundDir_Num, int foundSymlink_Num)
{
    /* delete old space */
    if (m_foundIO != NULL)
        delete m_foundIO;
    if (m_foundFile != NULL)
        delete m_foundFile;
    if (m_foundDir != NULL)
        delete m_foundDir;
    if (m_foundSymlink != NULL)
        delete m_foundSymlink;

    /* allocate new space */
    m_foundIO = new int [foundIO_Num];
    m_foundFile = new int [foundFile_Num];
    m_foundDir = new int [foundDir_Num];
    m_foundSymlink = new int [foundSymlink_Num];
}

void CClientFileSysRefThread::SetIdle()
{
    m_pool->SetIdleFileSysRefThread(m_id);
}

void CClientFileSysRefThread::SetSubdirAndStart(int subdir)
{
#ifndef WIN32
    LockMutex(m_lock);

    /* set sundir index of this run */
    m_subdir = subdir;

    m_ready = TRUE;
    pthread_cond_signal(&m_event);
    UnlockMutex(m_lock);
#else
    m_subdir = subdir;
    ReleaseSemaphore(m_readySem, 1, NULL);
#endif
}

void CClientFileSysRefThread::Quit()
{
    /* use subdir=-1 as quit signal */
    SetSubdirAndStart(-1);
}

void CClientFileSysRefThread::RefSubdir()
{
    int j, size;
    char dir_name[MAX_PATHLEN], name[MAX_PATHLEN], file_name[MAX_PATHLEN], target_name[MAX_PATHLEN];
    BOOL deleteTag;

    /* set all m_found tag as FALSE */
    for (j=0; j<m_IONum[m_subdir]; j++)
        m_foundIO[j] = FALSE;

    for (j=0; j<m_FileNum[m_subdir]; j++)
        m_foundFile[j] = FALSE;

    for (j=0; j<m_DirNum[m_subdir]; j++)
        m_foundDir[j] = FALSE;

    for (j=0; j<m_SymlinkNum[m_subdir]; j++)
        m_foundSymlink[j] = FALSE;

    /* get this sub directory name */
    m_fsbase->GetSubName(dir_name, m_subdir);

    /* open this sub directory */
#ifndef WIN32
    DIR *dirp = opendir(dir_name);
    if (dirp == NULL)
#else
    lstrcpy(file_name, dir_name);
    lstrcat(file_name, "\\*.*");
    WIN32_FIND_DATA FileData;
    HANDLE hSearch = FindFirstFile(file_name, &FileData);
    if (hSearch == INVALID_HANDLE_VALUE)
#endif
    {
        fprintf(stderr, "Make the sub direcotry (%d) first\n", m_subdir);
        return;
    }

    /* search this sub directory */
#ifndef WIN32
    struct dirent *dp;
    while ((dp = readdir(dirp)) != NULL)
    {
        strcpy(name, dp->d_name);
#else
    /* serarch this directory */
    do 
    {

        strcpy(name, FileData.cFileName);
#endif
        if (name[0] == '.')     /* ignore "." && ".." */
            continue;
        j = atoi(name+1);
        if (j == 0 && (strlen(name+1) != 6 || !isdigit(name[1])))
            j = -1;
        deleteTag = TRUE;

        if (name[0] == 'F' && (j >= 0 && j < m_IONum[m_subdir]))
        {
            deleteTag = FALSE;
            m_foundIO[j] = TRUE;
            m_fsbase->GetIOFileName(name, m_subdir, j);
            CSafeFileOperation::Lookup(name, &size, NULL);
            if (size > m_IOSize[m_subdir][j])
                CSafeFileOperation::Truncate(name, m_IOSize[m_subdir][j]);
            else if (size < m_IOSize[m_subdir][j])
                CSafeFileOperation::Append(name, m_IOSize[m_subdir][j] - size);
        }
        else if (name[0] == 'G' && (j >= 0 && j < m_FileNum[m_subdir]))
        {
            deleteTag = FALSE;
            m_foundFile[j] = TRUE;
            m_fsbase->GetNonIOFileName(name, m_subdir, j);
            CSafeFileOperation::Chmod(name, m_FileMode[m_subdir][j]);
        }
        else if (name[0] == 'D' && (j >= 0 && j < m_DirNum[m_subdir]))
        {
            deleteTag = FALSE;
            m_foundDir[j] = TRUE;
            m_DirMode[m_subdir][j/8] |= (1 << (j % 8));
        }
        else if (name[0] == 'L' && (j >= 0 && j < m_SymlinkNum[m_subdir]))
        {
            deleteTag = FALSE;
            m_foundSymlink[j] = TRUE;
            m_SymlinkMode[m_subdir][j/8] |= (1 << (j % 8));
        }

        if (deleteTag)
        {
            strcpy(file_name, dir_name);
            strcat(file_name, _STR_PATHSEP);
            strcat(file_name, name);
            CSafeFileOperation::RemoveAny(file_name);
        }
#ifndef WIN32
    }
    closedir(dirp);
#else
    } while (FindNextFile(hSearch, &FileData));
    FindClose(hSearch);
#endif

    /* create new IO files */
    for (j=0; j<m_IONum[m_subdir]; j++)
    {
        if (! m_foundIO[j])
        {
            m_fsbase->GetIOFileName(name, m_subdir, j);
            CSafeFileOperation::CreateWrite(name, m_IOSize[m_subdir][j]);
        }
    }

    /* create new NonIO files */
    for (j=0; j<m_FileNum[m_subdir]; j++)
    {
        if (! m_foundFile[j])
        {
            m_fsbase->GetNonIOFileName(name, m_subdir, j);
            CSafeFileOperation::CreateNull(name, m_FileMode[m_subdir][j]);
        }
    }

    /* create new dirs */
    for (j=0; j<m_DirNum[m_subdir]; j++)
    {
        if (! m_foundDir[j])
        {
            m_fsbase->GetDirName(name, m_subdir, j);
            CSafeFileOperation::Mkdir(name);
            m_DirMode[m_subdir][j/8] |= (1 << (j % 8));
        }
    }

    /* create new NonIO files */
    for (j=0; j<m_SymlinkNum[m_subdir]; j++)
    {
        if (! m_foundSymlink[j])
        {
            m_fsbase->GetSymlinkName(name, m_subdir, j);
            m_fsbase->GetSymlinkTargetName(target_name);
            CSafeFileOperation::Symlink(name, target_name);
            m_SymlinkMode[m_subdir][j/8] |= (1 << (j % 8));
        }
    }
}
