
/**************************************************************************
**
** Copyright (C) EMC Corporation 2001, All rights reserved.
**
** FILE NAME: Pumper.cpp
**
** DESCRIPTION: This is the main procedure of the pumper program.
**              
**
** AUTHOR: Yubing Wang       Mar-18-2001
**
** REVISION HISTORY:
** 
**  Who         Date       Description of Change
** ---------------------------------------------------------------------
**
** (This is only for the major code change that does not include bug fixes.
** Comments for bug fixes should be placed directly where the bugs are.)
**
****************************************************************************
**
**
** FUNCTION LIST:
**     int main(int argc, char *argv[])
**
**     int CPumper::AppMain(int argc, char *argv[]);
**     void CPumper::InitLoadGenThreadPool(int num);
**     void CPumper::WaitForRequest();
**
**
****************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#ifndef WIN32
#include <ctype.h>
#include <unistd.h>
#include <sys/mnttab.h>
#include <signal.h>
#include <time.h>
#endif
#include "Common.h"
#include "SocketServer.h"
#include "SocketClient.h"
#include "Packet.h"
#include "NetXDT.h"
#include "SysTime.h"
#include "ClientLoadGenPool.h"
#include "ClientFileSysThread.h"
#include "ClientPrimMgrThread.h"
#include "ClientLoadMgrThread.h"
#include "Pumper.h"
#include "ConfigFile.h"


#ifndef WIN32
#define USAGE       \
"pumper {-v | -p | -l | -stop} [file.cfg | prim_name]\n"\
"pumper -v                       --- display version number\n"\
"pumper -p [file.cfg]            --- start pumper as prim client\n"\
"pumper -l                       --- start pumper as load client\n"\
"pumper -stop [prim_name]        --- stop load\n"
#else
#define USAGE       \
"MPumper {-v | -p | -l | -stop} [file.cfg | prim_name]\n"\
"MPumper -v                       --- display version number\n"\
"MPumper -p [file.cfg]            --- start pumper as prim client\n"\
"MPumper -l                       --- start pumper as load client\n"\
"MPumper -stop [prim_name]        --- stop load\n"
#endif

#ifndef WIN32
CLocalPathManager theLocalPathManager;
#endif

CPumper thePumper;

#ifndef WIN32
extern "C"
{
    void broken_pipe(int)
    {
        printf("Broken Pipe Detected\n");
        signal(SIGPIPE, broken_pipe);
    }
}
#endif

int main(int argc, char *argv[])
{
    char ctime_buf[32];

    char config_file[1024];
    char prim_name[1024];

    /* command line option */
    int option = -1;
    if (argc > 1)
    {
        if (strcmp(argv[1], "-v") == 0)
        {
            if (argc == 2)
                option = 0;
        }
        else if (strcmp(argv[1], "-p") == 0)
        {
            option = 1;
            if (argc == 2)
                strcpy(config_file, "default.cfg");
            else if (argc == 3)
                strcpy(config_file, argv[2]);
            else
                option = 0;
        }
        else if (strcmp(argv[1], "-l") == 0)
        {
            if (argc == 2)
                option = 2;
        }
        else if (strcmp(argv[1], "-stop") == 0)
        {
            option = 3;
            if (argc == 2)
                strcpy(prim_name, "localhost");
            else if (argc == 3)
                strcpy(prim_name, argv[2]);
            else
                option = 0;
        }
    }

    /* error option */
    if (option == -1)
    {
        printf("Usage: %s\n", USAGE);
        return 0;
    }

    if (option == 0)
    {
        char date[256];
        strcpy(date, __DATE__);
        printf("Version 1.0. Built on %s\n", date);
        return 0;
    }

    printf("%s: Starting pumper ...\n", Safe_ctime(ctime_buf));
    fflush(stdout);

#ifdef WIN32
	WSADATA wsaData;
	if (WSAStartup(0x202, &wsaData) == SOCKET_ERROR)
	{
        printf("%s: WSAStartup failed with error %d\n", Safe_ctime(ctime_buf), WSAGetLastError());
        fflush(stdout);
		WSACleanup();
		return 1;
	}
#endif

#ifndef WIN32
    signal(SIGPIPE, broken_pipe);
#endif

    thePumper.AppMain(option, config_file, prim_name);

#ifdef WIN32
	WSACleanup();
#endif

    printf("%s: Pumper exit\n", Safe_ctime(ctime_buf));
    fflush(stdout);

    return 0;
}

/*************************** CThreadList ***************************/

CThreadList::CThreadList(int type)
{
    m_Type = type;
    m_ThreadNum = 0;
    m_MaxThreadNum = 32;
    m_Threads = new CBaseThread * [m_MaxThreadNum];
    InitMutex(m_lock);
}

CThreadList::~CThreadList()
{
    if (m_Threads != NULL)
        delete [] m_Threads;
#ifdef WIN32
    if (m_lock != NULL)
        CloseHandle(m_lock);
#endif
}

void CThreadList::AddThread(CBaseThread *thread)
{
    LockMutex(m_lock);

    if (m_ThreadNum == m_MaxThreadNum)
    {
        m_MaxThreadNum += 4;
        CBaseThread **old_Threads = m_Threads;
        m_Threads = new CBaseThread * [m_MaxThreadNum];
        for (int i=0; i<m_ThreadNum; i++)
            m_Threads[i] = old_Threads[i];
        delete [] old_Threads;
    }
    m_Threads[m_ThreadNum] = thread;
    m_ThreadNum ++;

    UnlockMutex(m_lock);
}

void CThreadList::DeleteThread(CBaseThread *thread)
{
    LockMutex(m_lock);

    int i;
    for (i=0; i<m_ThreadNum; i++)
    {
        if (m_Threads[i] == thread)
            break;
    }
    if (i < m_ThreadNum)
    {
        for (int j=i+1; j<m_ThreadNum; j++)
            m_Threads[j-1] = m_Threads[j];
        m_ThreadNum --;
    }

    UnlockMutex(m_lock);
}

CBaseThread *CThreadList::FindThreadByLoadID(int id)
{
    CBaseThread *thread = NULL;

    LockMutex(m_lock);

    /* 1:CClientFileSysThread 2:CClientPrimMgrThread, 3:CClientLoadMgrThread */
    int job_id, client_id;
    for (int i=0; i<m_ThreadNum; i++)
    {
        if (m_Type == 1)
        {
            ((CClientFileSysThread *) m_Threads[i])->GetID(job_id, client_id);
            if (job_id == id)
            {
                thread = m_Threads[i];
                break;
            }
        }
        else if (m_Type == 2)
        {
            if (((CClientPrimMgrThread *) m_Threads[i])->GetLoadID() == id)
            {
                thread = m_Threads[i];
                break;
            }
        }
        else
        {
            if (((CClientLoadMgrThread *) m_Threads[i])->GetLoadID() == id)
            {
                thread = m_Threads[i];
                break;
            }
        }
    }

    UnlockMutex(m_lock);

    return thread;
}

/************************ CLocalPathManager *************************/
/* this class is only for unix version */

#ifndef WIN32

#define LOCALPATH_PREFIX        "/LGFS/ffs"

CLocalPathManager::CLocalPathManager()
{
    InitMutex(m_lock);
    for (int i=0; i<MAX_CLIENT_FSNUM; i++)
        m_LocalPathFree[i] = TRUE;

    /* read the mount table to check the file system already mounted */
    FILE *fp = fopen("/etc/mnttab", "r");
    if (fp == NULL)
        return;

    struct mnttab mp;
    char *pIndex, *p;
    int index;
    while ( 1 )
    {
        if (getmntent(fp, &mp) != 0)
            break;
        if (strncmp(mp.mnt_mountp, LOCALPATH_PREFIX, strlen(LOCALPATH_PREFIX)) != 0)
            continue;

        /* pIndex should be a number */
        p = pIndex = mp.mnt_mountp + strlen(LOCALPATH_PREFIX);
        while (*p)
        {
            if (! isdigit(*p))
                break;
            p ++;
        }
        if (p != pIndex && ! (*p))
        {
            index = atoi(pIndex);
            if (index >= 0 && index < MAX_CLIENT_FSNUM)
                m_LocalPathFree[index] = FALSE;
        }
    }
    fclose(fp);
}

CLocalPathManager::~CLocalPathManager()
{
}

/* this function is only for UNIX client */
void CLocalPathManager::ApplyLocalPath(char *localpath)
{
    int i;
    LockMutex(m_lock);
    for (i=0; i<MAX_CLIENT_FSNUM; i++)
    {
        if (m_LocalPathFree[i])
            break;
    }
    _ASSERT(i < MAX_CLIENT_FSNUM);
    m_LocalPathFree[i] = FALSE;
    UnlockMutex(m_lock);
    sprintf(localpath, "%s%d", LOCALPATH_PREFIX, i);
}

/* this function is only for UNIX client */
void CLocalPathManager::FreeLocalPath(char *localpath)
{
    int i = atoi(localpath+strlen(LOCALPATH_PREFIX));
    LockMutex(m_lock);
    m_LocalPathFree[i] = TRUE;
    UnlockMutex(m_lock);
}

#endif

/**************************** CPumper ****************************/

CPumper::CPumper() : m_FileSysThreadList(1), m_PrimThreadList(2), m_LoadThreadList(3)
{
    opnum = -1;
//    if (argc == 2)
//        opnum = atoi(argv[1]);
}

CPumper::~CPumper()
{
}

int CPumper::AppMain(int option, char *config_file, char *prim_name)
{
    /* start pumper as prim client */
    if (option == 1)
    {
        CClientPrimMgrThread *prim_mgr_thread = new CClientPrimMgrThread(NULL);
        m_PrimThreadList.AddThread(prim_mgr_thread);
        prim_mgr_thread->SetConfigFile(config_file);
        prim_mgr_thread->Start();

        WaitForRequest();
    }

    /* start pumper as load client */
    if (option == 2)
    {
        WaitForRequest();
    }

    /* stop load */
    if (option == 3)
    {
        StopLoad(prim_name);
    }

    return 0;
}

void CPumper::WaitForRequest()
{
    char ctime_buf[32];
    CSocketServer req_sock;
    CSocketClient *msg_sock;
    CPacket packet;
    char *buffer;
    int size;
    int msgtype;

/*    char hostname[MAX_NAMELEN];
    if (gethostname(hostname, MAX_NAMELEN) == -1)
        strcpy(hostname, "UNKNOWN");*/

    if (! req_sock.AssignPort(CLIENT_MAIN_PORT))
    {
        printf("%s: Cannot assign the main port\n", Safe_ctime(ctime_buf));
        fflush(stdout);
        return;
    }

    while ( 1 )
	{
        printf("%s: Wait for request from resource allocator or prime client.\n", Safe_ctime(ctime_buf));
        fflush(stdout);

		msg_sock = req_sock.Wait();
        if (msg_sock == NULL)
            return;

        printf("%s: Request is coming. Start to read request.\n", Safe_ctime(ctime_buf));
        fflush(stdout);

        //msg_sock->SetNoDelayClose();

        /* read the message type firstly */
        if (! packet.Receive(msg_sock))
        {
            delete msg_sock;
            continue;
        }

        buffer = packet.GetData(&size);
        /* printf("%s: Got message: %d bytes\n", Safe_ctime(ctime_buf), size);
        fflush(stdout); */

        CNet_MsgType net_msgtype;
        net_msgtype.xdt_trans(XDTOP_GET, buffer, size);
        msgtype = net_msgtype.type;

        if (msgtype == NET_MSGTYPE_EXIT_SERVER)
        {
            delete msg_sock;
            break;
        }

        /* read the different message according to the message type */
        if (msgtype == NET_MSGTYPE_PUMPER_NONE)
        {
            printf("%s: Receive data of NET_MSGTYPE_PUMPER_NONE ...\n", Safe_ctime(ctime_buf));
            fflush(stdout);

            /* send back response with magic number */
            CNet_CommNum net_resp;
            net_resp.num = 0x1a2b3c4d;
            net_resp.xdt_trans(XDTOP_PUT, buffer, size);
            packet.SetData(buffer, size);
            packet.Send(msg_sock);
            delete msg_sock;
        }
        else if (msgtype == NET_MSGTYPE_PUMPER_START_REFFS)
        {
            printf("%s: Receive data of NET_MSGTYPE_PUMPER_START_REFFS ...\n", Safe_ctime(ctime_buf));
            fflush(stdout);

            /* generate the file system thread, it will read the message from the socket */
            CClientFileSysThread *fs_thread = new CClientFileSysThread(msg_sock);
            m_FileSysThreadList.AddThread(fs_thread);
            fs_thread->Start();
        }
        else if (msgtype == NET_MSGTYPE_PUMPER_STOP_REFFS)
        {
            printf("%s: Receive data of NET_MSGTYPE_PUMPER_STOP_REFFS ...\n", Safe_ctime(ctime_buf));
            fflush(stdout);

            /* it's from resource manager server, need to read job id */
            if (! packet.Receive(msg_sock))
                continue;
            buffer = packet.GetData(&size);
            CNet_LoadID net_loadid;
            net_loadid.xdt_trans(XDTOP_GET, buffer, size);

            CClientFileSysThread *thread = (CClientFileSysThread *) FindThreadByLoadID(1, net_loadid.id);
            if (thread != NULL)
                thread->SetBreak();

            delete msg_sock;
        }
        else if (msgtype == NET_MSGTYPE_PUMPER_START_PRIM)
        {
            /* generate the primary manager thread, it will read the message from the socket */
            CClientPrimMgrThread *prim_mgr_thread = new CClientPrimMgrThread(msg_sock);
//            prim_mgr_thread->SetHostName(hostname);
            m_PrimThreadList.AddThread(prim_mgr_thread);
            prim_mgr_thread->Start();
        }
        else if (msgtype == NET_MSGTYPE_PUMPER_STOP_PRIM)
        {
            printf("%s: Receive data of NET_MSGTYPE_PUMPER_STOP_PRIM ...\n", Safe_ctime(ctime_buf));
            fflush(stdout);

            /* it's from resource manager server, need to read load id */
            if (! packet.Receive(msg_sock))
                continue;
            buffer = packet.GetData(&size);
            CNet_LoadID net_loadid;
            net_loadid.xdt_trans(XDTOP_GET, buffer, size);

            CClientPrimMgrThread *thread = (CClientPrimMgrThread *) FindThreadByLoadID(2, net_loadid.id);
            if (thread != NULL)
            {
                /* read the reason (pause or stop) */
                if (! packet.Receive(msg_sock))
                    continue;
                buffer = packet.GetData(&size);
                CNet_StopLoadOnPrim net_stopload;
                net_stopload.xdt_trans(XDTOP_GET, buffer, size);

                if (thread->GetStatus() != 2)   // not refreshing
                    thread->SetSignal(net_stopload.reason);
                else
                    thread->SetRefSignal(net_stopload.reason+2);
            }

            delete msg_sock;
        }
        else if (msgtype == NET_MSGTYPE_PUMPER_START_LOAD)
        {
            /* generate the load manager thread, it will read the message from the socket */
            CClientLoadMgrThread *load_mgr_thread = new CClientLoadMgrThread(msg_sock);
//            load_mgr_thread->SetHostName(hostname);

            //load_mgr_thread->InitLoadGenThreadPool(DEFAULT_LOADGENNUM);
/*#ifndef WIN32
            load_mgr_thread->InitLoadGenThreadPool(16);
#else
            load_mgr_thread->InitLoadGenThreadPool(8);
#endif*/
            m_LoadThreadList.AddThread(load_mgr_thread);
            load_mgr_thread->Start();
        }
        else if (msgtype == NET_MSGTYPE_PUMPER_FSBASE)
        {
            printf("%s: Receive data of NET_MSGTYPE_PUMPER_FSBASE ...\n", Safe_ctime(ctime_buf));
            fflush(stdout);

            /* it's from primary client, need to read load id */
            if (! packet.Receive(msg_sock))
                continue;
            buffer = packet.GetData(&size);

            CNet_LoadID net_loadid;
            net_loadid.xdt_trans(XDTOP_GET, buffer, size);
            CClientLoadMgrThread *thread = (CClientLoadMgrThread *) FindThreadByLoadID(3, net_loadid.id);
            if (thread != NULL)
                thread->RecvFSBaseInfoFromPrim(msg_sock);
            delete msg_sock;
        }
        else if (msgtype == NET_MSGTYPE_PUMPER_FSINFO)
        {
            printf("%s: Receive data of NET_MSGTYPE_PUMPER_FSINFO ...\n", Safe_ctime(ctime_buf));
            fflush(stdout);

            /* it's from primary client, need to read load id */
            if (! packet.Receive(msg_sock))
                continue;
            buffer = packet.GetData(&size);

            CNet_LoadID net_loadid;
            net_loadid.xdt_trans(XDTOP_GET, buffer, size);
            CClientLoadMgrThread *thread = (CClientLoadMgrThread *) FindThreadByLoadID(3, net_loadid.id);
            if (thread != NULL)
            {
                if (thread->RecvFileSysInfoFromPrim(msg_sock))
                    thread->SetReady();
            }
            delete msg_sock;
        }
        else if (msgtype == NET_MSGTYPE_PUMPER_BREAK)
        {
            printf("%s: Receive data of NET_MSGTYPE_PUMPER_BREAK ...\n", Safe_ctime(ctime_buf));
            fflush(stdout);

            /* it's from primary client, need to read load id */
            if (! packet.Receive(msg_sock))
                continue;
            buffer = packet.GetData(&size);

            CNet_LoadID net_loadid;
            net_loadid.xdt_trans(XDTOP_GET, buffer, size);
            CClientLoadMgrThread *thread = (CClientLoadMgrThread *) FindThreadByLoadID(3, net_loadid.id);
            if (thread != NULL)
                thread->RecvBreakMessageFromPrim(msg_sock);
            delete msg_sock;
        }
        else if (msgtype == NET_MSGTYPE_PUMPER_GOAHEAD)
        {
            printf("%s: Receive data of NET_MSGTYPE_PUMPER_GOAHEAD ...\n", Safe_ctime(ctime_buf));
            fflush(stdout);

            /* it's from primary client, need to read load id */
            if (! packet.Receive(msg_sock))
                continue;
            buffer = packet.GetData(&size);

            CNet_LoadID net_loadid;
            net_loadid.xdt_trans(XDTOP_GET, buffer, size);
            CClientLoadMgrThread *thread = (CClientLoadMgrThread *) FindThreadByLoadID(3, net_loadid.id);
            if (thread != NULL)
            {
                if (thread->RecvGoAheadFromPrim(msg_sock, 20000))
                    thread->SetReady();
            }
            delete msg_sock;
        }
        else if (msgtype == NET_MSGTYPE_PUMPER_LOAD_ACK)
        {
            printf("%s: Receive data of NET_MSGTYPE_PUMPER_LOAD_ACK ...\n", Safe_ctime(ctime_buf));
            fflush(stdout);

            /* it's to primary client, need to read load id */
            if (! packet.Receive(msg_sock))
                continue;
            buffer = packet.GetData(&size);

            CNet_LoadID net_loadid;
            net_loadid.xdt_trans(XDTOP_GET, buffer, size);
            CClientPrimMgrThread *thread = (CClientPrimMgrThread *) FindThreadByLoadID(2, net_loadid.id);
            if (thread != NULL)
                thread->RecvAckFromClient(msg_sock);
            delete msg_sock;
        }
        else if (msgtype == NET_MSGTYPE_PUMPER_LOAD_FSINFO)
        {
            printf("%s: Receive data of NET_MSGTYPE_PUMPER_LOAD_FSINFO ...\n", Safe_ctime(ctime_buf));
            fflush(stdout);

            /* it's to primary client, need to read load id */
            if (! packet.Receive(msg_sock))
                continue;
            buffer = packet.GetData(&size);

            CNet_LoadID net_loadid;
            net_loadid.xdt_trans(XDTOP_GET, buffer, size);
            CClientPrimMgrThread *thread = (CClientPrimMgrThread *) FindThreadByLoadID(2, net_loadid.id);
            if (thread != NULL)
                thread->RecvFileSysInfoFromClient(msg_sock);
            delete msg_sock;
        }
        else if (msgtype == NET_MSGTYPE_PUMPER_REF_FSINFO)
        {
            printf("%s: Receive data of NET_MSGTYPE_PUMPER_REF_FSINFO ...\n", Safe_ctime(ctime_buf));
            fflush(stdout);

            /* it's to primary client, need to read load id */
            if (! packet.Receive(msg_sock))
                continue;
            buffer = packet.GetData(&size);

            CNet_LoadID net_loadid;
            net_loadid.xdt_trans(XDTOP_GET, buffer, size);
            CClientPrimMgrThread *thread = (CClientPrimMgrThread *) FindThreadByLoadID(2, net_loadid.id);
            if (thread != NULL)
            {
                thread->RecvRefFileSysInfoFromClient(msg_sock);
                thread->AddRefReady();
            }
            delete msg_sock;
        }
        else
        {
            printf("%s: Received an unknown message\n", Safe_ctime(ctime_buf));
            fflush(stdout);
            // ignore this unknown message and continue to receive next one
            delete msg_sock;
        }
	}
}

void CPumper::StopLoad(char *prim_name)
{
    char ctime_buf[32];

    printf("%s: Stop the load on %s.\n", Safe_ctime(ctime_buf), prim_name);
    fflush(stdout);

    /* connect to primary client's main port */
    CSocketClient socket;
    if (! socket.ConnectToServer(prim_name, CLIENT_MAIN_PORT))
    {
        printf("%s: Cannot connect to client %s.\n", Safe_ctime(ctime_buf), prim_name);
        fflush(stdout);
        return;
    }

    CPacket packet;
    char *buffer;
    int size;

    /* send the message type firstly */
    CNet_MsgType net_msgtype;
    net_msgtype.type = NET_MSGTYPE_PUMPER_STOP_PRIM;
    net_msgtype.xdt_trans(XDTOP_PUT, buffer, size);
    packet.SetData(buffer, size);
    if (! packet.Send(&socket))
    {
        printf("%s: Fail to send message (%d bytes) to client %s.\n", Safe_ctime(ctime_buf), size, prim_name);
        fflush(stdout);
        return;
    }

    /* then send the load id */
    CNet_LoadID net_loadid;
    net_loadid.id = 1;  /* always use 1 as load id */
    net_loadid.xdt_trans(XDTOP_PUT, buffer, size);
    packet.SetData(buffer, size);
    if (! packet.Send(&socket))
    {
        printf("%s: Fail to send message (%d bytes) to client %s.\n", Safe_ctime(ctime_buf), size, prim_name);
        fflush(stdout);
        return;
    }

    /* send the stop load command */
    CNet_StopLoadOnPrim net_stopload;
    net_stopload.reason = 2;    // stop
    net_stopload.xdt_trans(XDTOP_PUT, buffer, size);
    packet.SetData(buffer, size);
    if (! packet.Send(&socket))
    {
        printf("%s: Fail to send message (%d bytes) to client %s.\n", Safe_ctime(ctime_buf), size, prim_name);
        fflush(stdout);
        return;
    }

    printf("%s: Success to send stop command to client %s.\n", Safe_ctime(ctime_buf), prim_name);
    fflush(stdout);
}

void CPumper::AddThread(int type, CBaseThread *thread)
{
    if (type == 1)
        m_FileSysThreadList.AddThread(thread);
    else if (type == 2)
        m_PrimThreadList.AddThread(thread);
    else
        m_LoadThreadList.AddThread(thread);
}

void CPumper::DeleteThread(int type, CBaseThread *thread)
{
    if (type == 1)
        m_FileSysThreadList.DeleteThread(thread);
    else if (type == 2)
        m_PrimThreadList.DeleteThread(thread);
    else
        m_LoadThreadList.DeleteThread(thread);
}

CBaseThread *CPumper::FindThreadByLoadID(int type, int id)
{
    if (type == 1)
        return m_FileSysThreadList.FindThreadByLoadID(id);
    else if (type == 2)
        return m_PrimThreadList.FindThreadByLoadID(id);
    else
        return m_LoadThreadList.FindThreadByLoadID(id);
}


