
/**************************************************************************
**
** Copyright (C) EMC Corporation 2001, All rights reserved.
**
** FILE NAME: SafeFileOperation.cpp
**
** DESCRIPTION: This class defines a group of file operation.
**              
**
** AUTHOR: Yubing Wang       Mar-02-2001
**
** REVISION HISTORY:
** 
**  Who         Date       Description of Change
** ---------------------------------------------------------------------
**
** (This is only for the major code change that does not include bug fixes.
** Comments for bug fixes should be placed directly where the bugs are.)
**
****************************************************************************
**
**
** FUNCTION LIST:
**     CSafeFileOperation::CSafeFileOperation()
**     CSafeFileOperation::~CSafeFileOperation()
**
**
****************************************************************************/


/* define the initial buffer function for both Windows and Unix version */

#include "Common.h"
#include "SafeFileOperation.h"

#define WRITE_BUFFER_SIZE               8192

static BOOL init_tag = FALSE;
static char write_buffer[WRITE_BUFFER_SIZE];

static char *InitWriteBuffer()
{
    if (! init_tag)
	{
        uint *p = (uint *) write_buffer;
        uint *end = (uint *) &write_buffer[WRITE_BUFFER_SIZE];
        while (p < end)
        {
            *p = (uint) 0xbeefdead;
            p ++;
        }
        init_tag = TRUE;
    }
    return write_buffer;
}


#ifdef WIN32

/************************************************************/
/*                define the Windows version                */
/************************************************************/

/* return value:
   0:Fail  1: File exists  2: File doesn't exist */
int CSafeFileOperation::Lookup(char *name, int *size, int *mode)
{
    DWORD fileattr = GetFileAttributes(name);
    if (fileattr == 0xFFFFFFFF)     // file doesn't exist
    {
        DWORD errNo = GetLastError();
        if (errNo == ERROR_FILE_NOT_FOUND || errNo == ERROR_PATH_NOT_FOUND || errNo == ERROR_BAD_NETPATH)  // XX
            return 2;       /* file doesn't exist */
        else
            return 0;       /* failed */
    }

    if (mode != NULL)
        *mode = (int) fileattr;

    if (size != NULL)
    {
        DWORD filesize;
        if (fileattr & FILE_ATTRIBUTE_DIRECTORY)
            filesize = 0;
        else
        {
            HANDLE hFile = CreateFile(name,            // open filename 
                            GENERIC_READ,              // open for reading 
                            0,                         // don't share for reading 
                            NULL,                      // no security 
                            OPEN_EXISTING,             // existing file only 
                            FILE_ATTRIBUTE_NORMAL,     // normal file
                            NULL);                     // no attr. template
            filesize = GetFileSize(hFile, NULL);
            CloseHandle(hFile);
        }
        *size = (int) filesize;
    }
    return 1; /* file exists */
}

BOOL CSafeFileOperation::Chmod(char *name, ushort mode)
{
    return SetFileAttributes(name, (DWORD)mode);
}

BOOL CSafeFileOperation::CreateNull(char *name, ushort &mode)
{
    HANDLE hFile;                               // file HANDLE

    hFile = CreateFile(name,                    // open filename
                       GENERIC_WRITE,           // open for writing 
                       0,                       // don't share for reading 
                       NULL,                    // no security 
                       CREATE_ALWAYS,           // creating a new file 
                       FILE_ATTRIBUTE_NORMAL,   // normal file 
                       NULL);                   // no attr. template

    if (hFile == INVALID_HANDLE_VALUE)          // error
        return FALSE;

    CloseHandle(hFile);

    DWORD attr = FILE_ATTRIBUTE_ARCHIVE | FILE_ATTRIBUTE_HIDDEN | FILE_ATTRIBUTE_READONLY;
    SetFileAttributes(name, attr);
    mode = (ushort) attr;

    return TRUE;
}

BOOL CSafeFileOperation::Remove(char *name)
{
    SetFileAttributes(name, FILE_ATTRIBUTE_NORMAL);
    return DeleteFile(name);
}

BOOL CSafeFileOperation::Mkdir(char *name)
{
    return CreateDirectory(name, NULL);
}

BOOL CSafeFileOperation::Rmdir(char *name)
{
    return RemoveDirectory(name);
}

BOOL CSafeFileOperation::EmptyDir(char *name)
{
    /* open this direcotry */
    char filepath[MAX_PATHLEN];
    lstrcpy(filepath, name);
    lstrcat(filepath, "\\*.*");
    WIN32_FIND_DATA FileData;
    HANDLE hSearch = FindFirstFile(filepath, &FileData);
    if (hSearch == INVALID_HANDLE_VALUE)
        return FALSE;

    int mode;
    do 
    {
        if (strcmp(FileData.cFileName, ".") != 0 && strcmp(FileData.cFileName, "..") != 0)
        {
            lstrcpy(filepath, name);
            lstrcat(filepath, "\\");
            lstrcat(filepath, FileData.cFileName);
            if (Lookup(filepath, NULL, &mode) == 1)
            {
                if (mode & FILE_ATTRIBUTE_DIRECTORY)
                    RemoveDirectory(filepath);
                else
                    DeleteFile(filepath);
            }
        }
    } while (FindNextFile(hSearch, &FileData));

    FindClose(hSearch);

    return TRUE;
}

BOOL CSafeFileOperation::Symlink(char *name, char *target)
{
    return TRUE;
}

BOOL CSafeFileOperation::CreateWrite(char *name, int size)
{
    HANDLE hFile = CreateFile(name,                     // open filename
                             GENERIC_WRITE,             // open for writing 
                             0,                         // don't share for reading 
                             NULL,                      // no security 
                             CREATE_ALWAYS,             // open an existing file
                             FILE_ATTRIBUTE_NORMAL,     // normal file 
                             NULL);                     // no attr. template
    if (hFile == INVALID_HANDLE_VALUE)                  // file cann't open, error
        return FALSE;

    char *p = InitWriteBuffer();
    int wsize = 0;
    int blocksize = WRITE_BUFFER_SIZE;
    DWORD wsize1, wsize2;
    while (wsize < size)
    {
        wsize1 = (wsize + blocksize <= size) ? blocksize : (size - wsize);
        if (! WriteFile(hFile, p, wsize1, &wsize2, NULL))
        {
            CloseHandle(hFile);
            return FALSE;
        }
        wsize += wsize2;
    }

    CloseHandle(hFile);
    return TRUE;
}

BOOL CSafeFileOperation::Append(char *name, int size)
{
    HANDLE hFile = CreateFile(name,                     // open filename
                             GENERIC_WRITE,             // open for writing 
                             0,                         // don't share for reading 
                             NULL,                      // no security 
                             OPEN_EXISTING,             // open an existing file
                             FILE_ATTRIBUTE_NORMAL,     // normal file 
                             NULL);                     // no attr. template
    if (hFile == INVALID_HANDLE_VALUE)                  // file cann't open, error
        return FALSE;
    SetFilePointer(hFile, 0, NULL, FILE_END);

    char *p = InitWriteBuffer();
    int wsize = 0;
    int blocksize = WRITE_BUFFER_SIZE;
    DWORD wsize1, wsize2;
    while (wsize < size)
    {
        wsize1 = (wsize + blocksize <= size) ? blocksize : (size - wsize);
        if (! WriteFile(hFile, p, wsize1, &wsize2, NULL))
        {
            CloseHandle(hFile);
            return FALSE;
        }
        wsize += wsize2;
    }

    CloseHandle(hFile);
    return TRUE;
}

BOOL CSafeFileOperation::SetNull(char *name)
{
    HANDLE hFile = CreateFile(name,                    // open filename
                            GENERIC_WRITE,             // open for writing 
                            0,                         // don't share for reading 
                            NULL,                      // no security 
                            TRUNCATE_EXISTING,         // creating a new file 
                            FILE_ATTRIBUTE_NORMAL,     // normal file 
                            NULL);                     // no attr. template
    if (hFile == INVALID_HANDLE_VALUE)                 // error
        return FALSE;

    CloseHandle(hFile);
    return TRUE;
}

BOOL CSafeFileOperation::Truncate(char *name, int size)
{
    HANDLE hFile = CreateFile(name,                     // open filename
                             GENERIC_WRITE,             // open for writing 
                             0,                         // don't share for reading 
                             NULL,                      // no security 
                             OPEN_EXISTING,             // open an existing file
                             FILE_ATTRIBUTE_NORMAL,     // normal file 
                             NULL);                     // no attr. template
    if (hFile == INVALID_HANDLE_VALUE)                  // file cann't open, error
        return FALSE;

    if (SetFilePointer(hFile, (LONG)size, NULL, FILE_BEGIN) == INVALID_FILE_SIZE)
    {
        CloseHandle(hFile);
        return FALSE;
    }
    if (! SetEndOfFile(hFile))
        return -1;

    CloseHandle(hFile);
    return TRUE;
}

BOOL CSafeFileOperation::RemoveAny(char *name)
{
    int mode;
    int status = Lookup(name, NULL, &mode);
    if (status != 1)
        return FALSE;
    if (! (mode & FILE_ATTRIBUTE_DIRECTORY))
        return Remove(name);

    /* considered as a directory */
    char file_name[MAX_PATHLEN];
    HANDLE hSearch;
    WIN32_FIND_DATA FileData;

    /* empty this directory firstly */
    lstrcpy(file_name, name);
    lstrcat(file_name, "\\*.*");
    hSearch = FindFirstFile(file_name, &FileData);
    if (hSearch == INVALID_HANDLE_VALUE)
        return FALSE;

    do 
    {
        if (strcmp(FileData.cFileName, ".") != 0 && strcmp(FileData.cFileName, "..") != 0)
        {
            lstrcpy(file_name, name);
            lstrcat(file_name, "\\");
            lstrcat(file_name, FileData.cFileName);
            RemoveAny(file_name);
        }
    } while (FindNextFile(hSearch, &FileData));

    FindClose(hSearch);

    /* then delete this directory */
    return Rmdir(name);
}

BOOL CSafeFileOperation::CheckWritePermission(char *name)
{
    HANDLE hFile = CreateFile(name,                     // open filename
                             GENERIC_WRITE,             // open for writing 
                             0,                         // don't share for reading 
                             NULL,                      // no security 
                             CREATE_ALWAYS,             // open an existing file
                             FILE_ATTRIBUTE_NORMAL,     // normal file 
                             NULL);                     // no attr. template
    if (hFile == INVALID_HANDLE_VALUE)                  // file cann't open, error
        return FALSE;

    char bin_data[16];
    for (int i=0; i<16; i++)
        bin_data[i] = 0x50 + i;

    DWORD wsize;
    BOOL status = WriteFile(hFile, bin_data, 16, &wsize, NULL);
    CloseHandle(hFile);

    DeleteFile(name);

    return status && wsize == 16;
}


#else

/************************************************************/
/*                  define the Unix version                 */
/************************************************************/

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <dirent.h>
#include <errno.h>
#include "SafeFileOperation.h"


/* return value:
   0:Fail  1: File exists  2: File doesn't exist */
int CSafeFileOperation::Lookup(char *name, int *size, int *mode)
{
    struct stat statinfo;
    int status = lstat(name, &statinfo);
    if (status == -1)
    {
        if (errno == ENOENT)
            return 2;       /* file doesn't exist */
        else
            return 0;       /* failed */
    }

    if (mode != NULL)
        *mode = statinfo.st_mode;
    int filesize = (statinfo.st_mode & S_IFDIR) ? 0 : statinfo.st_size;
    if (size != NULL)
        *size = filesize;
    return 1; /* file exists */
}

BOOL CSafeFileOperation::Chmod(char *name, ushort mode)
{
    return chmod(name, mode) != -1;
}

BOOL CSafeFileOperation::CreateNull(char *name, ushort &mode)
{
    int fd = creat(name, 00644);
    mode = (ushort) 00644;
    if (fd < 0)
        return FALSE;
    close(fd);
    return TRUE;
}

BOOL CSafeFileOperation::Remove(char *name)
{
    return unlink(name) != -1;
}

BOOL CSafeFileOperation::Mkdir(char *name)
{
    return mkdir(name, 00755) != -1;
}

BOOL CSafeFileOperation::Rmdir(char *name)
{
    return rmdir(name) != -1;
}

BOOL CSafeFileOperation::EmptyDir(char *name)
{
    /* open this direcotry */
    DIR *dirp = opendir(name);
    if (dirp == NULL)
        return FALSE;

    /* delete all files in this direcotry */
    char filepath[MAX_PATHLEN];
    struct dirent *dp;
    int mode;
    while ((dp = readdir(dirp)) != NULL)
    {
        if (strcmp(dp->d_name, ".") == 0 || strcmp(dp->d_name, "..") == 0)
            continue;
        strcpy(filepath, name);
        strcat(filepath, "/");
        strcat(filepath, dp->d_name);
        if (Lookup(filepath, NULL, &mode) == 1)
        {
            if (mode & S_IFDIR)
                rmdir(filepath);
            else
                unlink(filepath);
        }
    }
    closedir(dirp);
    return TRUE;
}

BOOL CSafeFileOperation::Symlink(char *name, char *target)
{
    return symlink(target, name) != -1;
}

BOOL CSafeFileOperation::CreateWrite(char *name, int size)
{
    int fd = creat(name, 00644);
    if (fd < 0)
        return FALSE;

    char *p = InitWriteBuffer();
    int wsize = 0;
    int blocksize = WRITE_BUFFER_SIZE;
    int wsize1, wsize2;
    while (wsize < size)
    {
        wsize1 = (wsize + blocksize <= size) ? blocksize : (size - wsize);
        wsize2 = write(fd, p, wsize1);
        if (wsize2 == -1)
        {
            close(fd);
            return FALSE;
        }
        wsize += wsize2;
    }

    close(fd);
    return TRUE;
}

BOOL CSafeFileOperation::Append(char *name, int size)
{
    int fd = open(name, O_WRONLY | O_APPEND);
    if (fd < 0)
        return FALSE;

    char *p = InitWriteBuffer();
    int wsize = 0;
    int blocksize = WRITE_BUFFER_SIZE;
    int wsize1, wsize2;
    while (wsize < size)
    {
        wsize1 = (wsize + blocksize <= size) ? blocksize : (size - wsize);
        wsize2 = write(fd, p, wsize1);
        if (wsize2 == -1)
        {
            close(fd);
            return FALSE;
        }
        wsize += wsize2;
    }

    close(fd);
    return TRUE;
}

BOOL CSafeFileOperation::SetNull(char *name)
{
    return truncate(name, 0) == 0;
}

BOOL CSafeFileOperation::Truncate(char *name, int size)
{
    return truncate(name, size) == 0;
}

BOOL CSafeFileOperation::RemoveAny(char *name)
{
    int mode;
    int status = Lookup(name, NULL, &mode);
    if (status != 1)
        return FALSE;
    if (mode & S_IFREG)
        return Remove(name);

    /* considered as a directory */
    char file_name[MAX_PATHLEN];
    DIR *dirp;
    struct dirent *dp;

    /* empty this directory firstly */
    dirp = opendir(name);
    if (dirp == NULL)
        return FALSE;

    while ((dp = readdir(dirp)) != NULL)
    {
        if (strcmp(dp->d_name, ".") == 0 || strcmp(dp->d_name, "..") == 0)
            continue;
        strcpy(file_name, name);
        strcat(file_name, "/");
        strcat(file_name, dp->d_name);
        RemoveAny(file_name);
    }

    closedir(dirp);

    /* then delete this directory */
    return Rmdir(name);
}

BOOL CSafeFileOperation::CheckWritePermission(char *name)
{
    int fd = creat(name, 00666);
    if (fd < 0)
        return FALSE;

    char bin_data[16];
    for (int i=0; i<16; i++)
        bin_data[i] = 0x50 + i;
    int wsize = write(fd, bin_data, 16);
    close(fd);
    unlink(name);

    return wsize == 16;
}

#endif

